#include "OgreDepthShadowMapRenderingRun.h"
#include "OgreIlluminationManager.h"


OgreDepthShadowMapRenderingRun::OgreDepthShadowMapRenderingRun(OgreSharedRuns* sharedRuns,
																String name,
																Light* light,
																unsigned int resolutionX,
																unsigned int resolutionY,
																String materialName
														   )
														   :DepthShadowMapRenderingRun(resolutionX, resolutionY)
														   , OgreRenderingRun(1, 1)
														   , RenderingRun(1, 1)
{
	this->light = light;
	this->sharedRuns = sharedRuns;
	this->name = name;	
	
	this->materialName = materialName;

	createDepthMap();	
}

void OgreDepthShadowMapRenderingRun::createDepthMap()
{
	
		TexturePtr texPtr = Ogre::TextureManager::getSingleton().createManual(name, 
																			"default",
																			TEX_TYPE_2D,
																			resolutionX,
																			resolutionY,
																			0,
																			0,
																			PF_FLOAT16_RGBA,
																			TU_RENDERTARGET);
		depthMapTexture = texPtr.getPointer();
		depthMapCamera = Root::getSingleton()._getCurrentSceneManager()->createCamera(name + "_CAMERA");
		//add viewport to rendertarget
		HardwarePixelBuffer* hpb = (depthMapTexture->getBuffer()).getPointer();
		RenderTarget* rt = hpb->getRenderTarget();
		Viewport* v = rt->addViewport(depthMapCamera);
		v->setOverlaysEnabled(false);
		rt->setAutoUpdated(false);	
}

void OgreDepthShadowMapRenderingRun::updateFrame(unsigned long frameNum)
{
	refreshLight(frameNum);

	if(light->getType() == Light::LT_POINT)
	{
		updateDepthMap();
	}
	else
	{
		updateDepthMap();
	}
	
	
}

void OgreDepthShadowMapRenderingRun::updateDepthCubeFace(int facenum)
{
}

void OgreDepthShadowMapRenderingRun::updateDepthMap()
{
	setMaterialForVisibles(materialName, depthMapCamera, false);	
	RenderTarget* rt = depthMapTexture->getBuffer().getPointer()->getRenderTarget();	
	Viewport* vp = rt->getViewport(0);
	
	rt->update();
	restoreMaterials();
	rt->writeContentsToFile("shadowmap1.dds");
}

void setnear()
{

}

void OgreDepthShadowMapRenderingRun::refreshLight(unsigned long frameNum)
{
	
	if(light!= 0)
	{
		if(light->getType() == Light::LT_DIRECTIONAL)
		{
			Camera* viewcam = OgreIlluminationManager::getSingleton().getMainCamera();			
			
			OgreFocusingMapRenderingRun* frun = (OgreFocusingMapRenderingRun*)
				OgreIlluminationManager::getSingleton().getGlobalRun(ILLUMRUN_FOCUSING_MAP)->asOgreRenderingRun();
			frun->setCameraMatrices(viewcam->getViewMatrix(), viewcam->getProjectionMatrix());

			Vector3 lightpos = light->getPosition();
			Vector3 lightdir = light->getDirection();
			lightdir.normalise();		
			
			Vector3 min;
			Vector3 max;			

			bool uselispsm = true;
			if(uselispsm)
			{
				Camera* maincam = OgreIlluminationManager::getSingleton().getMainCamera();
				Vector3 viewdir = maincam->getDirection();
				viewdir.normalise();
				Vector3 campos = maincam->getPosition();
				float nearclip = maincam->getNearClipDistance();

				float dotprod = viewdir.dotProduct(lightdir);
				float sinGamma = sqrt(1.0 - dotprod * dotprod);
				
				Vector3 left = lightdir.crossProduct(viewdir);
				Vector3 up = left.crossProduct(lightdir);
				up.normalise();
				//left = lightdir.crossProduct(up);
				left.normalise();				

				Matrix4 viewMatrix( left.x,      left.y,      left.z,		 0,
									up.x,		 up.y,		  up.z,			 0,
									lightdir.x,  lightdir.y,  lightdir.z,    0,
									0,           0,           0,             1);				
				
				frun->setLightMatrix(viewMatrix);				
				OgreIlluminationManager::getSingleton().updateGlobalRun( ILLUMRUN_FOCUSING_MAP,
																			frameNum);			
				frun->getMinMax(min, max);

				float n;
				float f;
				float d = abs(max.y - min.y);
					
				float z_n = nearclip / sinGamma;
				float one_min_singamma = (1 - sinGamma);
				float fact = d * pow(one_min_singamma, 2);
				float z_f = z_n + d * sinGamma * fact;
				n = (z_n + sqrt(z_f * z_n)) / sinGamma;
				///n = 100000;
				f = n + d;

				Vector3 pos = campos - up * (n - nearclip);
				
				Matrix4 newViewMatrix(  left.x,      left.y,      left.z,		 -left.dotProduct(pos),
										up.x,		 up.y,		  up.z,			 -up.dotProduct(pos),
										lightdir.x,  lightdir.y,  lightdir.z,    -lightdir.dotProduct(pos),
										0,           0,           0,             1);


				Matrix4 newProjection = Matrix4::IDENTITY;
				
				newProjection[1][1] = (f + n) / (f - n);
				newProjection[1][3] = - 2* f * n / (f - n);
				newProjection[3][1] = 1;
				newProjection[3][3] = 0;
				//newProjection = newProjection.transpose();
				Matrix4 standardProjMatrix = Matrix4::IDENTITY;
				standardProjMatrix.setScale(Vector3(1,1,-1));
				newProjection = newProjection * standardProjMatrix;
				depthMapCamera->setCustomProjectionMatrix(true, newProjection);
				depthMapCamera->setCustomViewMatrix(true, viewMatrix);
			}
			else
			{
				Matrix4 standardProjMatrix = Matrix4::IDENTITY;
				standardProjMatrix.setScale(Vector3(0,0,-1));
				depthMapCamera->setCustomProjectionMatrix(true, standardProjMatrix);
				depthMapCamera->setCustomViewMatrix(false);
				depthMapCamera->setProjectionType(PT_ORTHOGRAPHIC);			
				depthMapCamera->setPosition(lightpos);
				depthMapCamera->setDirection(lightdir);
			}


			Matrix4 lightMatrix = depthMapCamera->getProjectionMatrix() * depthMapCamera->getViewMatrix();
			frun->setLightMatrix(lightMatrix);
			
			OgreIlluminationManager::getSingleton().updateGlobalRun( ILLUMRUN_FOCUSING_MAP,
																		frameNum);

			frun->getMinMax(min, max);			

						
			Vector3 scale = (max - min);
			scale *= 1.5; // just for safety
			scale.z = 200; //TODO: get scene bounding box to set z scale	
			Matrix4 projection = Matrix4::IDENTITY;
			projection.setScale(Vector3(2.0 / scale.x, 2.0 / scale.y, -2.0 / scale.z));			
			projection.setTrans(Vector3(-(max.x + min.x) / scale.x,
										-(max.y + min.y) / scale.y,
										-(max.z + min.z) / scale.z));
			projection = projection * depthMapCamera->getProjectionMatrix();
			depthMapCamera->setCustomProjectionMatrix(true, projection);
			
			
			
		}
		else if(light->getType() == Light::LT_SPOTLIGHT)
		{
	/*		Vector3 dir = light->getDirection();
			depthMapCamera->setDirection( dir );
			Vector3 pos = light->getParentNode()->getWorldPosition();
			depthMapCamera->setPosition(pos);
			depthMapCamera->setProjectionType(PT_PERSPECTIVE);		
			depthMapCamera->setFOVy(light->getSpotlightOuterAngle());*/
		}
		else//point light
		{
	/*		Vector3 pos = light->getParentNode()->getWorldPosition();
			Vector3 dir = -pos;
			depthMapCamera->setDirection( dir );
			depthMapCamera->setPosition(pos);
			depthMapCamera->setProjectionType(PT_PERSPECTIVE);
			depthMapCamera->setFOVy(Radian(Degree(140)));
			depthMapCamera->setNearClipDistance(1);
			// depthMapCamera->setFarClipDistance(200);
	
			//OGRE_EXCEPT(0, "NOT implemented for Pointlight", "OgreDepthShadowMapRenderingRun::refreshLight");
		*/	
			/*Vector3 pos = light->getParentNode()->getWorldPosition();
			Vector3 dir = -pos;
			depthMapCamera->setDirection( dir );
			depthMapCamera->setPosition(pos);
			depthMapCamera->setProjectionType(PT_ORTHOGRAPHIC);
			depthMapCamera->*/
			
		}
	}
}

Matrix4 OgreDepthShadowMapRenderingRun::getLightViewMatrix()
{
	return depthMapCamera->getViewMatrix();
}

Matrix4 OgreDepthShadowMapRenderingRun::getLightViewProjMatrix()
{
	return depthMapCamera->getProjectionMatrixWithRSDepth() * depthMapCamera->getViewMatrix();
}

